/*
 * Created on Tue Mar 07 2023
 *
 * Copyright (C) 2023 PixelApps - All Rights Reserved
 * The code in this file is Intellectutal Property of PixelApps
 * and can't be edited, redistributed or used without a valid license of it's product (Welltested)[https://welltested.ai]
 */
import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/constant/value.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:build/build.dart';
import 'package:source_gen/source_gen.dart';
import 'package:welltested/annotation.dart';
import 'package:welltested/src/executors/fixer_executor.dart';
import 'package:welltested/src/executors/models_executor.dart';
import 'package:welltested/src/executors/unit_test_executor.dart';
import 'package:welltested/src/executors/widget_test_executor.dart';
import 'package:welltested/src/executors/writer_executor.dart';
import 'package:welltested/src/extensions/class_element_extension.dart';
import 'package:welltested/src/models/test_request_data.dart';
import 'package:welltested/src/repositories/dart_server_repository.dart';
import 'package:welltested/src/repositories/generations_repository.dart';
import 'package:welltested/src/repositories/path_repository.dart';
import 'package:welltested/src/repositories/writer_repository.dart';
import 'package:welltested/src/utilities/utilities.dart';
import 'package:welltested/src/visitors/class_visitor.dart';

Future<String> generateWelltestedSource(
  BuildStep buildStep,
  ClassElement element, {
  required FixerExecutor fixerExecutor,
  required DartServerRepository dartServerRepository,
  ConstantReader? annotation, // TODO[Samyak] Can we make it `required`?
  bool save = false,
  required PathRepository pathRepository,
}) async {
  final List<String> excludedMethods = annotation
          ?.read('excludedMethods')
          .listValue
          .map((e) => e.toStringValue() ?? "")
          .toList() ??
      [];

  final List<DartObject> _rawScreenTypes =
      annotation?.read('screenTypes').listValue ?? [];
  final List<ScreenType> screenTypes = [];
  for (final type in _rawScreenTypes) {
    final name = type.getField("_name")?.toStringValue();
    if (name != null) {
      screenTypes.add(ScreenType.values.firstWhere((t) => t.name == name));
    }
  }

  final classAstNode =
      (await buildStep.resolver.astNodeFor(element)) as ClassDeclaration;
  final fileAstNode = classAstNode.parent;
  final fileImport =
      LibraryReader(pathRepository.inputLibrary).pathToElement(element);
  final GenerationRepository generationsRepository = GenerationRepository();
  final ModelsExecutor modelsExecutor = ModelsExecutor();
  final WriterRepository writerRepository = WriterRepository(buildStep);
  final WriterExecutor writerExecutor = WriterExecutor(
      pathRepository, writerRepository, dartServerRepository, fixerExecutor);

  final widgets = classAstNode.checkIfWidget();

  final WidgetTestExecutor widgetTestExecutor =
      WidgetTestExecutor(generationsRepository, pathRepository);
  final UnitTestExecutor unitTestExecutor =
      UnitTestExecutor(generationsRepository, pathRepository, modelsExecutor);

  try {
    if (widgets != null && widgets.isNotEmpty) {
      final request = await widgetTestExecutor.prepareRequest(
          widgets, element, screenTypes);

      final testTypes = ['rendering', 'behaviour'];
      for (int i = 0; i < testTypes.length; i++) {
        final testType = testTypes[i];
        if (save) {
          final file = await pathRepository.correspondingPermanentTestFile(
              "${testType}_" + request.className.toLowerCase());
          if (file != null) {
            await generationsRepository.updateFixedTestWidget(
                request, await file.readAsString(), testType);
          }
        } else {
          final widgetTests =
              await generationsRepository.buildWidgetTests(request);
          final overrideResult =
              widgetTests['${testType}_gen']['fixed_completion'];
          if (overrideResult != null) {
            await writerExecutor.outputGeneration(
                i,
                "${testType}_" + request.className.toLowerCase(),
                overrideResult);
            continue;
          }
          await writerExecutor.processGeneration(
              request.fileImportPath,
              widgetTests['${testType}_gen']['model_completion'],
              "${testType}_" + request.className.toLowerCase(),
              i);
        }
      }
    } else {
      final classVisitor = ClassVisitor();
      element.visitChildren(classVisitor);

      final Map<String, MethodDeclaration> methodAstNodes = {};
      for (final method in classVisitor.methods.entries) {
        final node =
            (await getResolvedAstNodeForElement(method.value, buildStep))
                as MethodDeclaration;
        methodAstNodes[method.key] = node;
      }

      for (int i = 0; i < methodAstNodes.entries.length; i++) {
        final method = methodAstNodes.entries.elementAt(i);

        if (excludedMethods.contains(method.key) || isPrivate(method.key))
          continue; // Don't generate separate test for excluded and private methods

        final TestRequestData testRequestData =
            await unitTestExecutor.prepareRequest(
                method: method,
                fileAstNode: fileAstNode!,
                fileImport: fileImport,
                className: classVisitor.className,
                methodAstNodes: methodAstNodes,
                buildStep: buildStep);

        if (save) {
          final file =
              await pathRepository.correspondingPermanentTestFile(method.key);
          if (file == null) {
            continue;
          }

          await generationsRepository.updateFixedTest(
              testRequestData, await file.readAsString());
        } else {
          log.info(" Generating Test for ${testRequestData.functionName}");
          final unitTests =
              await generationsRepository.buildUnitTests(testRequestData);

          final overrideResult = unitTests['fixed_completion'];
          if (overrideResult != null) {
            final file =
                await pathRepository.correspondingPermanentTestFile(method.key);
            if (file != null && (file.readAsString() == overrideResult)) {
              continue;
            }
            await writerExecutor.outputGeneration(
                i, method.key, overrideResult);
            continue;
          }
          if (unitTests['model_completion'] == null) {
            log.warning(
                "⛔️ Sorry, we can't process ${testRequestData.className} at the moment. Please try again Later.");
            continue;
          }
          await writerExecutor.processGeneration(testRequestData.fileImportPath,
              unitTests['model_completion'], method.key, i);
        }
      }
    }
  } catch (e) {
    throw (e);
  } finally {
    dartServerRepository.dispose();
  }

  return '';
}
